#include "allegro.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>

#include <hill.h>
#include <main.h>
#include <jumper.h>
#include <replay.h>
#include <api.h>
#include <draw.h>

double lowest_point_y;
int hill[MAX_POINTS][3];
int inrun[MAX_POINTS][3];
double hillrec_x;

double cache[CACHE_SIZE];
double in_cache[CACHE_SIZE];
char cache_ready;
int cache_count;


inline void q3d_later(  double x1, double y1, double z1,
                double x2, double y2, double z2,
                double x3, double y3, double z3,
                double x4, double y4, double z4, int color){
        V3D_f _p[4], _pout[50], _ptmp[50];
        V3D_f *p[4], *pout[50], *ptmp[50];
        int some_tmp[50];
        int i,n;
        for (i=0;i<4;++i)
                p[i]=&_p[i];
        for (i=0;i<8;++i){
                pout[i]=&_pout[i];
                ptmp[i]=&_ptmp[i];
        }
        p[0]->x=x1;
        p[0]->y=y1;
        p[0]->z=z1;
        p[0]->u=0;
        p[0]->v=0;
        p[1]->x=x2;
        p[1]->y=y2;
        p[1]->u=0;
        p[1]->v=32;
        p[1]->z=z2;
        p[2]->x=x3;
        p[2]->y=y3;
        p[2]->z=z3;
        p[2]->u=32;
        p[2]->v=32;
        p[3]->x=x4;
        p[3]->y=y4;
        p[3]->z=z4;
        p[3]->u=32;
        p[3]->v=0;
        for (i=0;i<4;++i)
                apply_matrix_f(&camera,p[i]->x,p[i]->y,p[i]->z,&p[i]->x,&p[i]->y,&p[i]->z);
        n=clip3d_f(POLYTYPE_FLAT,0.1,fog_dist,4,(AL_CONST V3D_f **)p,pout,ptmp,some_tmp);
        if (n<=0) return;
        for (i=0;i<n;++i)
                persp_project_f(pout[i]->x, pout[i]->y, pout[i]->z, &pout[i]->x, &pout[i]->y);
        if (polygon_z_normal_f(pout[2], pout[1], pout[0]) < 0) return;
        pout[0]->c=color;
        if (scene_polygon3d_f(POLYTYPE_FLAT | POLYTYPE_ZBUF, NULL, n, pout)){
                message("scene_polygon3d_f: failed");
                exit(1);
        }
}

double lagrange(double x, int surface[][3]){		//cached version of lagrange (speeds things up)
	if (x<-1000) return cache[0];
	if (surface==hill)
		return cache[(int)x+1000];
	return in_cache[(int)x+1000];
}
double lagrange_real(double x, int surface[][3]){            //Lagrange's approximation algorithm
	int i,j;
	int n_points=surface[0][0];
	double sum=0,num,den;
	for (i=1;i<=n_points;++i){
		num=den=1;
		for (j=1;j<=n_points;++j)
			if (j!=i){
				num*=x-surface[j][0];
				den*=surface[i][0]-surface[j][0];
			}
		sum+=surface[i][1]*num/den;
	}
	if (lowest_point>1)
		if (x>lowest_point) return (sum/FLATTEN+lowest_point_y*(FLATTEN-1)/FLATTEN);
	return sum;
}
int maxv(int a, int b){
	return a>b ? a : b;
}
int minv(int a, int b){
	return a<b ? a : b;
}
template <class T>
T min(T a, T b){
	return a<b ? a: b;
}
template <class T>
T max(T a, T b){
	return a>b ? a: b;
}
double hill_w(double x){
	return min(max((x-inrun[INRUN_N][0])/15+(INRUN_W),(INRUN_W)),1000.0);
}

int slope(double x){
	return (int)(50*(lagrange(x+1,hill)-lagrange(x,hill)));
}
int inslope(double x){
	return (int)(50*(lagrange(x+1,inrun)-lagrange(x,inrun)));
}
double norm(double xp, double dist){
	double a;
	double x=xp;//>offsetx+30 ? xp : offsetx+30;
	a=(inrun[INRUN_N][0]-x)/dist;
	a=ceil(a);
	return inrun[INRUN_N][0]-a*dist;
}

// The following functions give more than 50% increase of frame rate compared to the old ones.
// I'm quite happy about this fact ;).

void hill_floor(int base){	//base = how many points the curve is divided into
	double x,dist;
	int i;
	// the ground
	i=0;	
	dist=(double)(SCREEN_W+2*fog_dist)/(base-1);
	for (x=norm(jumper.pxd-fog_dist,dist);i<2*base;x+=dist){
		q3d(	x,lagrange(x,hill),-15*hill_w(lowest_point),
			x,lagrange(x,hill),15*hill_w(lowest_point),
			x-dist,lagrange(x-dist,hill),15*hill_w(lowest_point),
			x-dist,lagrange(x-dist,hill),-15*hill_w(lowest_point),
			darken(makecol(220,220,220),-slope(x)));

				
		i+=2;
	}
//	if (option(O_AUDIENCE)){
	audience(inrun[INRUN_N][0],1000,-2*hill_w(lowest_point),makecol(100,100,230));
	audience(lowest_point-300,450,-2*hill_w(lowest_point),makecol(100,100,230));
	audience(lowest_point+300,450,-2*hill_w(lowest_point),makecol(230,100,100));
	audience(lowest_point+900,250,-2*hill_w(lowest_point),makecol(50,230,50));
	audience(lowest_point+1300,600,-2*hill_w(lowest_point),makecol(50,150,230));
//	}


}

void inrun_floor(int base){	 
	double x,dist;
	int i;
	dist=(double)(inrun[INRUN_N][0]-inrun[1][0])/(base-1);
	x=norm(inrun[1][0]-dist+5,dist);
	q3d(	x+300,lagrange(x+300,inrun)+80,-INRUN_W-20,
		x+300,lagrange(x+300,hill),-INRUN_W-20,
		x+300,lagrange(x+300,hill),INRUN_W+20,
		x+300,lagrange(x+300,inrun)+80,INRUN_W+20,
		darken(INRUN_C,-20));
	
	dist=(double)(inrun[INRUN_N][0]-inrun[1][0])/(base-1);
	x=norm(inrun[INRUN_N][0],dist);
	q3d(	x-300,lagrange(x-300,inrun)+80,-INRUN_W-20,
		x-300,lagrange(x-300,inrun)+80,INRUN_W+20,
		x-300,lagrange(x-300,hill),INRUN_W+20,
		x-300,lagrange(x-300,hill),-INRUN_W-20,
		darken(INRUN_C,-20));
	dist=(double)(inrun[INRUN_N][0]-inrun[1][0])/(base-1);
	x=norm(inrun[INRUN_N][0],dist);
	q3d(	x,lagrange(x,inrun),-INRUN_W-20,
		x,lagrange(x,hill),-INRUN_W-20,
		x,lagrange(x,hill),INRUN_W+20,
		x,lagrange(x,inrun),INRUN_W+20,
		darken(INRUN_C,0));



	


// the inrun surface
	
	i=0;
	dist=(double)(inrun[INRUN_N][0]-inrun[1][0])/(base-1);
	if (dist>0) {
		for (x=norm(inrun[1][0]+5,dist);i<2*base;x+=dist){
			q3d(	x,lagrange(x,inrun),-INRUN_W,
				x,lagrange(x,inrun),INRUN_W,
				x-dist,lagrange(x-dist,inrun),INRUN_W,
				x-dist,lagrange(x-dist,inrun),-INRUN_W,
				makecol(255-abs(inslope(x)),255-abs(inslope(x)),255-abs(inslope(x))));
			i+=2;
		}
	}


//trails

	i=0;
	dist=(double)(inrun[INRUN_N][0]-inrun[1][0])/(base-1);
	if (dist>0) {
		for (x=norm(inrun[1][0]+5,dist);i<2*base;x+=dist){
			q3d(	x,lagrange(x,inrun),-SKI_SPACE*JUMPER_SIZE-5,
				x,lagrange(x,inrun),-SKI_SPACE*JUMPER_SIZE,
				x-dist,lagrange(x-dist,inrun),-SKI_SPACE*JUMPER_SIZE,
				x-dist,lagrange(x-dist,inrun),-SKI_SPACE*JUMPER_SIZE-5,
				makecol(220-inslope(x),220-inslope(x),220-inslope(x)));
			i+=2;
		}
	}
	
	i=0;
	dist=(double)(inrun[INRUN_N][0]-inrun[1][0])/(base-1);
	if (dist>0) {
		for (x=norm(inrun[1][0]+5,dist);i<2*base;x+=dist){
			q3d(	x,lagrange(x,inrun),SKI_SPACE*JUMPER_SIZE,
				x,lagrange(x,inrun),SKI_SPACE*JUMPER_SIZE+5,
				x-dist,lagrange(x-dist,inrun),SKI_SPACE*JUMPER_SIZE+5,
				x-dist,lagrange(x-dist,inrun),SKI_SPACE*JUMPER_SIZE,
				makecol(220-inslope(x),220-inslope(x),220-inslope(x)));
			i+=2;
		}
	}
	
	dist=(double)(inrun[INRUN_N][0]-inrun[1][0])/(base-1);
	x=norm(inrun[1][0]-dist+5,dist);
	q3d(	x,lagrange(x,inrun)-20,-INRUN_W,
		x,lagrange(x,inrun),-INRUN_W,
		x,lagrange(x,inrun),INRUN_W,
		x,lagrange(x,inrun)-20,INRUN_W,
		INRUN_C);

	
// the vertical rectangle at the end of the inrun
	
/*	points[0]=(int)m3dx(inrun[INRUN_N][0]-offsetx,INRUN_W);
	points[1]=(int)m3dy(inrun[INRUN_N][1]-offsety,INRUN_W);
	points[2]=(int)m3dx(inrun[INRUN_N][0]-offsetx,-INRUN_W);
	points[3]=(int)m3dy(inrun[INRUN_N][1]-offsety,-INRUN_W);
	points[4]=(int)m3dx(inrun[INRUN_N][0]-offsetx,-INRUN_W);
	points[5]=(int)m3dy(inrun[INRUN_N][1]-offsety+70,-INRUN_W);
	points[6]=(int)m3dx(inrun[INRUN_N][0]-offsetx,INRUN_W);
	points[7]=(int)m3dy(inrun[INRUN_N][1]-offsety+70,INRUN_W);*/
/*	if (points[2]>points[0])
	polygon(screen_bufor,4,points,INRUN_C-100); else
		line(screen_bufor,points[0],points[1],points[2],points[3],65535);*/

// k-point
	
	q3d(	k_point,lagrange(k_point,hill),-hill_w((double)k_point),
		k_point+10,lagrange(k_point+10,hill),-hill_w((double)k_point),
		k_point+10,lagrange(k_point+10,hill),hill_w((double)k_point),
		k_point,lagrange(k_point,hill),hill_w((double)k_point),
		makecol(255,0,0));

	

// hill record

	q3d(	hillrec_x,lagrange(hillrec_x,hill),-hill_w(hillrec_x),
		hillrec_x+10,lagrange(hillrec_x+10,hill),-hill_w(hillrec_x),
		hillrec_x+10,lagrange(hillrec_x+10,hill),hill_w(hillrec_x),
		hillrec_x,lagrange(hillrec_x,hill),hill_w(hillrec_x),
		makecol(0,0,255));



	
}
			// the inrun left side
void inrun_left(int base){
	double dist,x;
	int i;
	i=0;
	dist=(double)(inrun[INRUN_N][0]-inrun[1][0])/(base-1);
	if (dist>0) {
		for (x=norm(inrun[1][0]+5,dist);i<2*base;x+=dist){
			q3d(	x,lagrange(x,inrun)-20,-INRUN_W,
				x,lagrange(x,inrun),-INRUN_W,
				x-dist,lagrange(x-dist,inrun),-INRUN_W,
				x-dist,lagrange(x-dist,inrun)-20,-INRUN_W,
				INRUN_C);
			
			i+=2;
		}
	}
	i=0;
	dist=(double)(inrun[INRUN_N][0]-inrun[1][0])/(base-1);
	if (dist>0) {
		for (x=norm(inrun[1][0]+5,dist);i<2*base;x+=dist){
			q3d(	x,lagrange(x,inrun)+80,-INRUN_W-20,
				x,lagrange(x,inrun)-20,-INRUN_W-20,
				x-dist,lagrange(x-dist,inrun)-20,-INRUN_W-20,
				x-dist,lagrange(x-dist,inrun)+80,-INRUN_W-20,
				INRUN_C);
			
			i+=2;
		}
	}	

	i=0;
	dist=(double)(inrun[INRUN_N][0]-inrun[1][0])/(base-1);
	if (dist>0) {
		for (x=norm(inrun[1][0]+5,dist);i<2*base;x+=dist){
			q3d(	x,lagrange(x,inrun)-20,-INRUN_W-20,
				x,lagrange(x,inrun)-20,-INRUN_W,
				x-dist,lagrange(x-dist,inrun)-20,-INRUN_W,
				x-dist,lagrange(x-dist,inrun)-20,-INRUN_W-20,
				INRUN_C+100);
			i+=2;
		}
	}
	
	dist=(double)(inrun[INRUN_N][0]-inrun[1][0])/(base-1);
	x=norm(inrun[1][0]-dist+5,dist);
	q3d(	x,lagrange(x,inrun)+80,-INRUN_W-20,
		x,lagrange(x,hill),-INRUN_W-20,
		x+300,lagrange(x+300,hill),-INRUN_W-20,
		x+300,lagrange(x+300,inrun)+80,-INRUN_W-20,
		INRUN_C);
	
	dist=(double)(inrun[INRUN_N][0]-inrun[1][0])/(base-1);
	x=norm(inrun[INRUN_N][0],dist);
	q3d(	x,lagrange(x,inrun)+80,-INRUN_W-20,
		x-300,lagrange(x-300,inrun)+80,-INRUN_W-20,
		x-300,lagrange(x-300,hill),-INRUN_W-20,
		x,lagrange(x,hill),-INRUN_W-20,
		INRUN_C);





}
void hill_left(int base){
	int i;
	double dist,x;
	i=0;
	dist=(double)(SCREEN_W+2*fog_dist)/(base-1);
	if (dist>0) {
		for (x=norm(max(jumper.pxd-fog_dist,(double)(inrun[INRUN_N][0]+200)+dist),dist);i<2*base;x+=dist){
			q3d(	x,lagrange(x,hill),-hill_w(x),
				x-dist,lagrange(x-dist,hill),-hill_w(x-dist),
				x-dist,lagrange(x-dist,hill)-30,-hill_w(x-dist)-15,
				x,lagrange(x,hill)-30,-hill_w(x)-15,
				makecol(100,255,100)+(int)(100*x));
			
			q3d(	x,lagrange(x,hill),-hill_w(x)-30,
				x,lagrange(x,hill)-30,-hill_w(x)-15,
				x-dist,lagrange(x-dist,hill)-30,-hill_w(x-dist)-15,
				x-dist,lagrange(x-dist,hill),-hill_w(x-dist)-30,
				makecol(100,255,100)+(int)(100*x));
			
			i+=2;
		}
	}

	
}
					//right side
void inrun_right(int base){
	double dist,x;
	int i;
	i=0;
	dist=(double)(inrun[INRUN_N][0]-inrun[1][0])/(base-1);	
	if (dist>0) {
		for (x=norm(inrun[1][0]+5,dist);i<2*base;x+=dist){
			q3d(	x,lagrange(x,inrun)-20,INRUN_W+20,
				x,lagrange(x,inrun)+80,INRUN_W+20,
				x-dist,lagrange(x-dist,inrun)+80,INRUN_W+20,
				x-dist,lagrange(x-dist,inrun)-20,INRUN_W+20,
				INRUN_C);
			i+=2;
		}
	}
	i=0;
	dist=(double)(inrun[INRUN_N][0]-inrun[1][0])/(base-1);
	if (dist>0) {
		for (x=norm(inrun[1][0]+5,dist);i<2*base;x+=dist){
			q3d(	x,lagrange(x,inrun),INRUN_W,
				x,lagrange(x,inrun)-20,INRUN_W,
				x-dist,lagrange(x-dist,inrun)-20,INRUN_W,
				x-dist,lagrange(x-dist,inrun),INRUN_W,
				INRUN_C);
			i+=2;
		}
	}

	i=0;
	dist=(double)(inrun[INRUN_N][0]-inrun[1][0])/(base-1);
	if (dist>0) {
		for (x=norm(inrun[1][0]+5,dist);i<2*base;x+=dist){
			q3d(	x,lagrange(x,inrun)-20,INRUN_W,
				x,lagrange(x,inrun)-20,INRUN_W+20,
				x-dist,lagrange(x-dist,inrun)-20,INRUN_W+20,
				x-dist,lagrange(x-dist,inrun)-20,INRUN_W,
				INRUN_C+100);
			i+=2;
		}
	}
	dist=(double)(inrun[INRUN_N][0]-inrun[1][0])/(base-1);
	x=norm(inrun[1][0]-dist+5,dist);
	q3d(	x,lagrange(x,inrun)+80,INRUN_W+20,
		x+300,lagrange(x+300,inrun)+80,INRUN_W+20,
		x+300,lagrange(x+300,hill),INRUN_W+20,
		x,lagrange(x,hill),INRUN_W+20,

		INRUN_C);
	
	dist=(double)(inrun[INRUN_N][0]-inrun[1][0])/(base-1);
	x=norm(inrun[INRUN_N][0],dist);
	q3d(	x-300,lagrange(x-300,inrun)+80,INRUN_W+20,
		x,lagrange(x,inrun)+80,INRUN_W+20,
		x,lagrange(x,hill),INRUN_W+20,
		x-300,lagrange(x-300,hill),INRUN_W+20,

		INRUN_C);




}
void hill_right(int base){
	int i;
	double dist,x;
	i=0;
	dist=(double)(SCREEN_W+2*fog_dist)/(base-1);
	if (dist>0){
		for (x=norm(max(jumper.pxd-fog_dist,(double)(inrun[INRUN_N][0]+200)+dist),dist);i<2*base;x+=dist){
			q3d(	x,lagrange(x,hill),hill_w(x),
				x,lagrange(x,hill)-30,hill_w(x)+15,
				x-dist,lagrange(x-dist,hill)-30,hill_w(x-dist)+15,
				x-dist,lagrange(x-dist,hill),hill_w(x-dist),

				makecol(100,255,100)+(int)(100*x));
			
			q3d(	x,lagrange(x,hill),hill_w(x)+30,
				x-dist,lagrange(x-dist,hill),hill_w(x-dist)+30,
				x-dist,lagrange(x-dist,hill)-30,hill_w(x-dist)+15,
				x,lagrange(x,hill)-30,hill_w(x)+15,


				makecol(100,255,100)+(int)(100*x));

			i+=2;
		}
	}


}


// The rest

void hill_ceil(int base){
	double dist, x;
	int i;
	dist=(double)(inrun[INRUN_N][0]-inrun[1][0])/(base-1);
	x=norm(inrun[1][0]-dist+5,dist);
	q3d(	x,lagrange(x,hill),-INRUN_W-20,
		x,lagrange(x,inrun)-20,-INRUN_W-20,
		x,lagrange(x,inrun)-20,INRUN_W+20,
		x,lagrange(x,hill),INRUN_W+20,
		INRUN_C);
	


	i=0;
	dist=(double)(inrun[INRUN_N][0]-inrun[1][0])/(base-1);
	if (dist>0) {
		for (x=norm(inrun[1][0]+5,dist);i<2*base;x+=dist){
			q3d(	x,lagrange(x,inrun)+79,-INRUN_W-20,
				x-dist,lagrange(x-dist,inrun)+79,-INRUN_W-20,
				x-dist,lagrange(x-dist,inrun)+79,INRUN_W+20,
				x,lagrange(x,inrun)+79,INRUN_W+20,
				INRUN_C);
			i+=2;
		}
	}

/*	dist=(double)(inrun[INRUN_N][0]-inrun[1][0])/(base-1);
	x=norm(inrun[1][0]-dist,dist);
	q3d(	x-offsetx,lagrange(x,inrun)-offsety-20,-INRUN_W,
		x-offsetx,lagrange(x,inrun)-offsety,-INRUN_W,
		x-offsetx,lagrange(x,inrun)-offsety,INRUN_W,
		x-offsetx,lagrange(x,inrun)-offsety-20,INRUN_W,
		INRUN_C);*/



#ifndef DEBUG
	if (jumper.where_is==GROUND){
/*		textprintf_centre(screen_bufor,font,SCREEN_W/2,50,H_TEXT_COLOR,"result: %.2f m",result/ZOOM);
		textprintf_centre(screen_bufor,font,SCREEN_W/2,65,H_TEXT_COLOR,"press <SPACE> to jump again, <ESC> to go back or <R> to see the replay.");*/
		if (result>hill_rec*ZOOM && jumper.landing_type==OK && game_type==COMPETITION)
			textprintf_centre(screen_bufor,font,SCREEN_W/2,80,H_TEXT_COLOR,"new record!");
	}
	textprintf_right(screen_bufor,font,SCREEN_W,SCREEN_H-8,0,"Open Ski Jumping %s / Michal Brzozowski",VERSION);
	
	textprintf(screen_bufor,font,2,SCREEN_H-8,0,"%s K%d",hill_name,k_dist);
	rectfill(screen_bufor,SCREEN_W-WIND_REC_W-WIND_REC_X-3,WIND_REC_Y-8,SCREEN_W-WIND_REC_X+3,WIND_REC_Y+WIND_REC_H+3,0);
	rectfill(screen_bufor,SCREEN_W-WIND_REC_W-WIND_REC_X,WIND_REC_Y-5,SCREEN_W-WIND_REC_X,WIND_REC_Y+WIND_REC_H,65535);
	draw_egg(SCREEN_W-WIND_REC_X-WIND_REC_W/2,
		 WIND_REC_Y+WIND_REC_H/2,
	 	 SCREEN_W-WIND_REC_X-WIND_REC_W/2-wind,
		 WIND_REC_Y+WIND_REC_H/2+side_wind/5,5,0);
	draw_egg(SCREEN_W-WIND_REC_X-WIND_REC_W/2-wind,
		 WIND_REC_Y+WIND_REC_H/2+side_wind/5,
		 SCREEN_W-WIND_REC_X-WIND_REC_W/2-wind+(wind<0 ? -5 : 5),
		 WIND_REC_Y+WIND_REC_H/2+5+side_wind/5,5,0);
	draw_egg(SCREEN_W-WIND_REC_X-WIND_REC_W/2-wind,
		 WIND_REC_Y+WIND_REC_H/2+side_wind/5,	
		 SCREEN_W-WIND_REC_X-WIND_REC_W/2-wind+(wind<0 ? -5 : 5),
		 WIND_REC_Y+WIND_REC_H/2-5+side_wind/5,5,0);
	textprintf(screen_bufor,font,SCREEN_W-WIND_REC_X-WIND_REC_W+2,WIND_REC_Y-3,0,"Wind Measure");

#endif


}
void hill_bench(){
/*	int points[8];
	points[0]=(int)m3dx(inrun[0][0]-offsetx+ZOOM,INRUN_W);
	points[1]=(int)m3dy(lagrange(inrun[0][0]+ZOOM,inrun)-offsety-30,INRUN_W);
	points[2]=(int)m3dx(inrun[0][0]-offsetx+ZOOM+10,INRUN_W);
	points[3]=(int)m3dy(lagrange(inrun[0][0]+ZOOM,inrun)-offsety-20,INRUN_W);
	points[4]=(int)m3dx(inrun[0][0]-offsetx+ZOOM+10,-INRUN_W);
	points[5]=(int)m3dy(lagrange(inrun[0][0]+ZOOM,inrun)-offsety-20,-INRUN_W);
	points[6]=(int)m3dx(inrun[0][0]-offsetx+ZOOM,-INRUN_W);
	points[7]=(int)m3dy(lagrange(inrun[0][0]+ZOOM,inrun)-offsety-30,-INRUN_W);
	polygon(screen_bufor,4,points,INRUN_C-100);*/
//	q3d(inrun[0][0]-offsetx+ZOOM,inrun[0][1]-30,INRUN_W,
	q3d_later(	ZOOM+7,lagrange(ZOOM,inrun)-20,INRUN_W,
		ZOOM-10,lagrange(ZOOM,inrun)-20,INRUN_W,
		ZOOM-10,lagrange(ZOOM,inrun)-20,-INRUN_W,
		ZOOM+7,lagrange(ZOOM,inrun)-20,-INRUN_W,makecol(150,100,30));

}
void hill_small(int x1, int y1, int l, int h, int a, int b, int c, int d){
	int i;
	
	drawing_mode(DRAW_MODE_TRANS,NULL,0,0);
	set_trans_blender(0,0,0,128);	
	
	rectfill(screen_bufor,x1,y1,x1+l,y1+h,makecol(60,60,100));
	drawing_mode(DRAW_MODE_SOLID,NULL,0,0);
	
	for (i=a;i<b;++i)
		if ((lagrange(i,hill)-c)*h/(d-c)<h-2 && (lagrange(i,hill)-c)*h/(d-c)>0)
		vline(screen_bufor,x1+(i-a)*l/(b-a),y1+(int)(lagrange(i,hill)-c)*h/(d-c),
				y1+(int)(lagrange(i,hill)-c)*h/(d-c)+2,0);
	for (i=a;i<b;++i)
		if ((lagrange(i,inrun)-c)*h/(d-c)<h-2 && (lagrange(i,inrun)-c)*h/(d-c)>0)
		if (i<inrun[INRUN_N][0])
			vline(screen_bufor,x1+(i-a)*l/(b-a),y1+(int)(lagrange(i,inrun)-c)*h/(d-c),
				y1+(int)(lagrange(i,inrun)-c)*h/(d-c)+2,0);
	
			
}
double distance(double x){
	double sum=0;
	double i;
	double space=ZOOM/150.0;
	i=inrun[INRUN_N][0];
	while (i<x){
		sum+=sqrt(space*space+
				(lagrange_real(i+space,hill)-lagrange_real(i,hill))*
				(lagrange_real(i+space,hill)-lagrange_real(i,hill)));
		i+=space;
	}
	return sum;
}
void read_hill(char *n){
        FILE *in;
	char fullpath[100];
        int i;
	sprintf(fullpath,"%s%s",DATADIR,n);
	in=fopen(fullpath,"r");
	if (!in){
		message("error: file not found: %s%s",DATADIR,n);
		exit(1);
	}
	fscanf(in,"%s%d",hill_name,&k_dist);
        fscanf(in,"%d%d",&inrun[0][0],&hill[0][0]);
        for (i=1;i<=inrun[0][0];++i){
	                fscanf(in,"%d%d%d",&inrun[i][0],&inrun[i][1],&inrun[i][2]);
	                inrun[i][0]=(int)(inrun[i][0]*HILL_SIZE);
	                inrun[i][1]=(int)(inrun[i][1]*HILL_SIZE);
	        }
        for (i=1;i<=hill[0][0];++i){
	                fscanf(in,"%d%d%d",&hill[i][0],&hill[i][1],&hill[i][2]);
	                hill[i][0]=(int)(hill[i][0]*HILL_SIZE);
	                hill[i][1]=(int)(hill[i][1]*HILL_SIZE);
	        }
        fclose(in);
	lowest_point=0;
	lowest_point=find_lowest_point();
	lowest_point_y=lagrange_real(lowest_point,hill);


	k_point=(int)find_k(k_dist*ZOOM);
	hill_rec=get_record(n);
	hillrec_x=find_k(hill_rec*ZOOM);
	make_cache();
//	message("%f %f",hill_rec,hill_rec/ZOOM);
}
double find_k(double k){	// binary search (it's not necessary, but doesn't take much time anyway).
	double x=(inrun[INRUN_N][0]+hill[HILL_N][0])/2;
	double y=(x-inrun[INRUN_N][0])/2;
	while (y>5){
		if (distance(x)<k)
			x+=y; else
				x-=y;
		y*=0.5;
	}
	return x;
}
double find_lowest_point(){	// another binary search (this one is necessary).
	double x=(inrun[INRUN_N][0]+hill[HILL_N][0])/2;
	double y=(x-inrun[INRUN_N][0])/2;
	while (y>5){
		if (lagrange_real(x+5,hill)>lagrange_real(x,hill))
			x+=y; else
				x-=y;
		y*=0.5;
	}
	return x;
}
void new_record(char *c, double dist, char *name){
	FILE *rec=fopen(CFGDIR"hillrec.txt","r");
	FILE *hills=fopen(CFGDIR"hills.txt","r");
	double records[255];
	char buf[255];
	char names[255][255];
	int i=0,j;
//	message("new record: %.2lf %s",dist,name);
	if (!rec){
		message("error: file not found: hillrec.txt");
		exit(1);
	}
	if (!hills){
		message("error: file not found: hills.txt");
		exit(1);
	}
	while (!feof(rec)){
		size_t len;
		char *buf1=new char[255];
		char *c;
		int k;
		fgets(buf1,255,rec);
		if (feof(rec)) break;
		if (!isdigit(buf1[0])) break;
		sscanf(buf1,"%lf",&records[i]);
		c=strchr(buf1,' ');
		++c;
		k=strlen(c);
		while (k && isspace(c[k-1]) ) --k;
		strncpy(names[i],c,k);
		names[i][k]=0;
//		message("%d %s .",k,names[i]);
		++i;
		delete buf1;
	}
	for (j=0;j<i;++j){
		fscanf(hills,"%s",buf);
		if (!strcmp(buf,c)){
			records[j]=dist;
			strcpy(names[j],name);
		}
	}
	fclose(rec);
	fclose(hills);
	rec=fopen(CFGDIR"hillrec.txt","w");
	for (j=0;j<i;++j)
		fprintf(rec,"%lf %s\n",records[j],names[j]);
	fclose(rec);
}
double get_record(char *c){
	FILE *rec=fopen(CFGDIR"hillrec.txt","r");
	FILE *hill=fopen(CFGDIR"hills.txt","r");
	char buf[100];
	double dist;
//	message("get record %s",c);
	if (!rec){
		message("error: file not found: hillrec.txt");
		exit(1);
	}
	if (!hill){
		message("error: file not found: hills.txt");
		exit(1);
	}
	while (fscanf(hill,"%s",buf)!=EOF){
		char buf2[255];
		fgets(buf2,255,rec);
		sscanf(buf2,"%lf",&dist);
		if (!strcmp(buf,c)){
//			message("%s %lf",c,dist);
//			message("done");
			return dist;
		}
	}
	return 0;
}
void make_cache(){
	int x;
	cache_ready=0;
	cache_count=0;
	for (x=0;x<CACHE_SIZE;++x){
		cache[x]=lagrange_real((double)x-1000,hill);
		in_cache[x]=lagrange_real((double)x-1000,inrun);
	}
	cache_ready=1;
}
void show_records(){
	FILE *hills=fopen(CFGDIR"hills.txt","r");
	FILE *rec=fopen(CFGDIR"hillrec.txt","r");
	FILE *hill;
	char *hillfile[255];
	char *hillname[255];
	char buf[255];
	double record;
	int n=0;
	int choice;
	char name[255];
	menu_t rmenu=menu_t("Hill records");
	if (!rec){
		message("error: file not found: hillrec.txt");
		exit(1);
	}
	if (!hills){
		message("error: file not found: %shills.txt",CFGDIR);
		exit(1);
	}
	while (!feof(hills)){
		hillfile[n]=new char[255];
		hillname[n]=new char[255];
		if (fscanf(hills,"%s",hillfile[n])==EOF) break;
		sprintf(buf,"%s%s",DATADIR,hillfile[n]);
		hill=fopen(buf,"r");
		if (!hill){
			message("error: file not found: %s",buf);
			exit(1);
		}
		fscanf(hill,"%s",hillname[n]);
		char *buf1=new char[255];
		size_t len;
		fgets(buf1,255,rec);
		sscanf(buf1,"%lf",&record);
		strcpy(name,strchr(buf1,' ')+1);
		rmenu.add_item("%-25s%.2lf m,  %s",hillname[n],record,name);
		fclose(hill);
		++n;
	}
	fclose(rec);
	fclose(hills);
	rmenu.add_item("Reset all records");
	rmenu.add_item("Back");
	n+=2;
	while ((choice=rmenu.get_choice())!=n-1){
		if (choice<=n-3){
			menu_t cmenu=menu_t("Reset this record?");
			cmenu.add_item("No");
			cmenu.add_item("Yes");
			if (cmenu.get_choice()==1){
				new_record(hillfile[choice],0,"empty");
				rmenu.change_item(choice,"%-25s%.2lf m,  %s",hillname[choice],0.0,"empty");
			}
		}
		if (choice==n-2){
			menu_t cmenu=menu_t("Reset all records?");
			cmenu.add_item("No");
			cmenu.add_item("Yes");
			if (cmenu.get_choice()==1){
				for (int i=0;i<n-2;++i){
					new_record(hillfile[i],0.0,"empty");
					rmenu.change_item(i,"%-25s %.2lf m,  %s",hillname[i],0.0,"empty");
				}
			}
		}
	}
	for (int i=0;i<n-2;++i){
		delete hillfile[i];
		delete hillname[i];
	}
}
